import { Point } from "@influxdata/influxdb-client";
import { Request, Response } from "express";
import Cache from "node-cache";
import { Route, RouteMessages } from "server";
import { InfluxWriteApi, nodeManager } from "..";
import { MetricsSchema } from "../db/metrics";

const IGNORED_PATHS = ["/favicon.ico"];

/**
 * The total requests that have been made
 * TODO: move this to a metrics file
 */
let totalRequests: number | undefined;

const cache = new Cache({
  stdTTL: 300, // 5 minutes
});
type CachedRequest = {
  nodeId: string;
  status: number;
  headers: any;
  data: any;
};

/**
 * Logs a request
 *
 * @param nodeId the node ID that handled the request
 * @param url the URL of the request
 * @param status the status code of the request
 * @param time the time it took to handle the request (or true if it was cached)
 */
function log(
  nodeId: string,
  url: string,
  status: number,
  time: number | boolean
) {
  console.log(
    `[node: ${nodeId}] ${url} - ${status} (${
      time === true ? "cached" : time + "ms"
    })`
  );
}

type InfluxLog = {
  nodeId: string;
  url: string;
  status: number;
  time?: number;
  cached?: boolean;
};
/**
 * Creates an InfluxDB point and writes it to the database
 *
 * @param nodeId the node ID that handled the request
 * @param url the URL of the request
 * @param status the status code of the request
 * @param time the time it took to handle the request (or true if it was cached)
 */
async function logRequestToDatabase({
  nodeId,
  url,
  status,
  time,
  cached,
}: InfluxLog) {
  totalRequests = totalRequests ? totalRequests + 1 : 1;
  MetricsSchema.updateOne(
    { _id: "proxy" },
    { $set: { totalRequests: totalRequests } },
    { upsert: true }
  ).exec();

  const point = new Point("proxy");
  point.tag("type", "request");
  point.tag("node", nodeId);
  point.stringField("url", url);
  point.intField("status", status);
  if (cached) {
    point.tag("cached", "true");
  } else {
    point.intField("time", time as number);
  }
  try {
    InfluxWriteApi.writePoint(point);
    InfluxWriteApi.writePoint(
      new Point("proxy").intField("totalRequests", totalRequests)
    );
  } catch (ex) {
    console.log("Failed to write to influx");
    console.log(ex);
  }
}

export default class ProxyRoute extends Route {
  constructor() {
    super({ path: "/" });
  }

  async handle(req: Request, res: Response) {
    const url = req.url.substring(1);
    if (!url) {
      res.status(400).json(RouteMessages.badRequest("No URL provided"));
      return;
    }
    if (IGNORED_PATHS.includes(url)) {
      res.status(404).send("Not found");
      return;
    }

    const node = nodeManager.getRandomNode();
    if (!node) {
      res
        .status(500)
        .json(RouteMessages.internalServerError("No nodes available"));
      return;
    }

    try {
      const cachedRequest = cache.get<CachedRequest>(url);
      if (cachedRequest) {
        res
          .status(cachedRequest.status)
          .set(cachedRequest.headers)
          .json(cachedRequest.data);
        log(cachedRequest.nodeId, url, cachedRequest.status, true);
        logRequestToDatabase({
          nodeId: cachedRequest.nodeId,
          url,
          status: cachedRequest.status,
          cached: true,
        });
        return;
      }

      const before = Date.now();
      const response = await node.fetch(url);
      const nodeId = response.headers["x-proxy-node"];
      const data = response.data;

      if (response.status === 500) {
        res.status(500).json(RouteMessages.internalServerError(data));
        return;
      }

      log(nodeId, url, response.status, Date.now() - before);

      // Return the JSON response
      res.status(response.status).set(response.headers).json(data);
      cache.set(url, {
        nodeId: nodeId,
        status: response.status,
        headers: response.headers,
        data: data,
      } as CachedRequest);
      logRequestToDatabase({
        nodeId,
        url,
        status: response.status,
        time: Date.now() - before,
      });
    } catch (ex: any) {
      res.status(500).json(RouteMessages.internalServerError(ex.message || ex));
    }
  }
}
